@echo off

rem test Microsoft SQL Server against full test suite
echo Removes various databases!!
pause

call dotsql test
if errorlevel 1 goto error
call dotsql bank
if errorlevel 1 goto error
call dotsql employee
if errorlevel 1 goto error
call dotsql sales
if errorlevel 1 goto error
call dotsql supplier
if errorlevel 1 goto error

goto end

:error
echo One of the test scripts failed.
pause
goto end

:end
